<!-- NOTE: It is safe to edit this file. -->
<!-- This file is only initially generated by aac gen-plugin, and it won't be overwritten if the file already exists. -->
<!-- Be sure to remove these comments as this file will be used for publishing to site documentation -->
{% block before_style -%}{%- endblock before_style %}
# {{ plugin.name }}

{{ plugin.description }}

{%- if plugin.commands | length > 0 -%}
{% set optionals = [] %} {# Variable used for keeping track of whether we have already displayed the 'Optional Arguments' header #}
{%- for command in plugin.commands %}
## {{ command.name }} Command
{% if command.input | length > 0 %}
### Arguments
    {%- for input in command.input %}
    {% if not input.name.startswith('--') -%}{{ input.name }} {{"("}}{{input.type}}{{")"}} {# If an argument does not start with '--' it is mandatory #}
        {{ input.description }}
    {%- else %} {# else the input.name.startswith('--') so it is an Optional Argument #}
{%- if not optionals[0] %} {# Check to see if this is the first Optional Argument #}
### Optional Arguments
{%- set temp = optionals.append(true) -%} {# Set this flag so we don't write the 'Optional Arguments' header again for this command #}
{%- endif %}
    {{input.name}} {{"("}}{{input.type}}{{")"}}
        {{ input.description -}}
{%- endif -%}
{%- endfor -%}
{%- if optionals[0] is defined -%} {# Check that this is defined before popping. This covers cases with no Optional Arguments #}
{%- set temp = optionals.pop(0) -%} {# Pop the optional arguments flag so we can be ready for the next command if there is one #}
{%- endif -%}
{% else %} {# account for plugins with no arguments #}
### Arguments
    There are no arguments for the {{ command.name }} command
{%- endif %}

{{"<Insert command screenshot here>"}}
{% endfor %}
{%- endif %}
## Help

{{"<Insert help screenshot here>"}}

## Examples

{{"For viewing example input and output of executing this command, view <Insert link to command execution examples here>"}}
