# NOTE: It is safe to edit this file.
# This file is only initially generated by aac gen-project, and it won't be overwritten if the file already exists.

[testenv]
deps = .[test]
commands = py39: python -m unittest

[flake8]
extend-select = W504
ignore = E501, E402, D202, D401, W503, D212
exclude = .git,__pycache__,docs/source/conf.py,old,build,dist,venv,.env,.eggs,.tox
max-complexity = 10
max-line-length = 99
statistics = True
show-source = True
docstring-convention = google
count = True
per-file-ignores =
    # Ignore docstring errors in tests
    *tests/*:D102,D100,D101,D200,ANN101,ANN201,W293,W291,ANN001,E731

[unittest]
start-dir = .
code-directories = src

[coverage]
coverage = src
always-on = True
coverage-config = tox.ini

[run]
omit =
  *.jinja2

[report]
fail_under = 80.00
precision = 2
show_missing = True
skip_empty = True
sort = cover

[html]
directory = html_code_coverage
skip_covered = True

[tox]
isolated_build = True
skipsdist = True
envlist =
  py39
